@@ -3,30 +3,34 @@ BEGIN { require 5.006 }
 use strict;
 use warnings;
 
-our $VERSION = '0.003002';
+our $VERSION = '0.003003';
 $VERSION = eval $VERSION;
 
 use Exporter (); BEGIN { *import = \&Exporter::import }
 our @EXPORT_OK = qw(module_content module_handle);
 our %EXPORT_TAGS = (all => [@EXPORT_OK]);
 
-use File::Spec;
+use File::Spec ();
 use Scalar::Util qw(reftype refaddr openhandle);
-use Carp;
+use Carp qw(croak);
 use Config ();
 use Errno qw(EACCES);
-use constant _PMC_ENABLED => !(
-  exists &Config::non_bincompat_options ? grep { $_ eq 'PERL_DISABLE_PMC' } Config::non_bincompat_options()
-  : $Config::Config{ccflags} =~ /(?:^|\s)-DPERL_DISABLE_PMC\b/
+use constant _OPEN_LAYERS     => "$]" >= 5.008_000 ? ':' : '';
+use constant _ABORT_ON_EACCES => "$]" >= 5.017_001;
+use constant _ALLOW_PREFIX    => "$]" >= 5.008009;
+use constant _VMS             => $^O eq 'VMS' && !!require VMS::Filespec;
+use constant _WIN32           => $^O eq 'MSWin32';
+use constant _PMC_ENABLED     => !(
+  exists &Config::non_bincompat_options
+    ? grep { $_ eq 'PERL_DISABLE_PMC' } Config::non_bincompat_options()
+    : $Config::Config{ccflags} =~ /(?:^|\s)-DPERL_DISABLE_PMC\b/
 );
-use constant _VMS => $^O eq 'VMS' && !!require VMS::Filespec;
-use constant _WIN32 => $^O eq 'MSWin32';
 use constant _FAKE_FILE_FORMAT => do {
-  (my $uvx = $Config::Config{uvxformat}||'') =~ tr/"\0//d;
+  my $uvx = $Config::Config{uvxformat} || '';
+  $uvx =~ tr/"\0//d;
   $uvx ||= 'lx';
   "/loader/0x%$uvx/%s"
 };
-use constant _OPEN_LAYERS => "$]" >= 5.008 ? ':' : '';
 
 sub _mod_to_file {
   my $module = shift;
@@ -70,6 +74,10 @@ sub new {
     if !exists $options{pmc};
   $options{open} = 1
     if !exists $options{open};
+  $options{abort_on_eacces} = _ABORT_ON_EACCES
+    if !exists $options{abort_on_eacces};
+  $options{check_hooks_for_nonsearchable} = 1
+    if !exists $options{check_hooks_for_nonsearchable};
   bless \%options, $class;
 }
 
@@ -104,13 +112,6 @@ sub _searchable {
 sub _find {
   my ($self, $file, $all) = @_;
 
-  if (!_searchable($file)) {
-    my $open = $self->_open_file($file);
-    return $open
-      if $open;
-    croak "Can't locate $file";
-  }
-
   my @found;
   eval {
     if (my $found = $self->{found}) {
@@ -128,9 +129,30 @@ sub _find {
     die $@
       if $@;
   }
+
+  my $searchable = _searchable($file);
+  if (!$searchable) {
+    my $open = $self->_open_file($file);
+    if ($all) {
+      push @found, $open;
+    }
+    elsif ($open) {
+      return $open;
+    }
+    else {
+      croak "Can't locate $file";
+    }
+  }
+
   my $search = $self->{inc};
   for my $inc (@$search) {
     my $open;
+    if (!$searchable) {
+      last
+        if !$self->{check_hooks_for_nonsearchable};
+      next
+        if !length ref $inc;
+    }
     eval {
       if (!length ref $inc) {
         my $full = _VMS ? VMS::Filespec::unixpath($inc) : $inc;
@@ -165,21 +187,23 @@ sub _open_file {
     $full,
   ) {
     my $pmc = $full ne $try;
-    next
-      if -e $try ? (-d _ || -b _) : $! != EACCES;
-
-    if (!$self->{open} ? -e _ : open my $fh, '<'._OPEN_LAYERS, $try) {
-      return Module::Reader::File->new(
-        filename        => $file,
-        ($fh ? (raw_filehandle => $fh) : ()),
-        found_file      => $full,
-        disk_file       => $try,
-        is_pmc          => $pmc,
-        (defined $inc ? (inc_entry => $inc) : ()),
-      );
+    if (-e $try) {
+      next
+        if -d _ || -b _;
+      if (open my $fh, '<'._OPEN_LAYERS, $try) {
+        return Module::Reader::File->new(
+          filename        => $file,
+          ($self->{open} ? (raw_filehandle => $fh) : ()),
+          found_file      => $full,
+          disk_file       => $try,
+          is_pmc          => $pmc,
+          (defined $inc ? (inc_entry => $inc) : ()),
+        );
+      }
     }
+
     croak "Can't locate $file:   $full: $!"
-      unless $pmc;
+      if $self->{abort_on_eacces} && $! == EACCES && !$pmc;
   }
   return;
 }
@@ -205,10 +229,15 @@ sub _open_ref {
   my $fake_file = sprintf _FAKE_FILE_FORMAT, refaddr($inc), $file;
 
   my $fh;
+  my $prefix;
   my $cb;
   my $cb_options;
 
-  if (reftype $cb[0] eq 'GLOB' && openhandle $cb[0]) {
+  if (_ALLOW_PREFIX && reftype $cb[0] eq 'SCALAR') {
+    $prefix = shift @cb;
+  }
+
+  if ((reftype $cb[0]||'') eq 'GLOB' && openhandle $cb[0]) {
     $fh = shift @cb;
   }
 
@@ -217,13 +246,14 @@ sub _open_ref {
     # only one or zero callback options will be passed
     $cb_options = @cb > 1 ? [ $cb[1] ] : undef;
   }
-  elsif (!$fh) {
+  elsif (!defined $fh && !defined $prefix) {
     return;
   }
   return Module::Reader::File->new(
     filename => $file,
     found_file => $fake_file,
     inc_entry => $inc,
+    (defined $prefix ? (prefix => $prefix) : ()),
     (defined $fh ? (raw_filehandle => $fh) : ()),
     (defined $cb ? (read_callback => $cb) : ()),
     (defined $cb_options ? (read_callback_options => $cb_options) : ()),
@@ -232,12 +262,12 @@ sub _open_ref {
 
 sub inc   { $_[0]->{inc} }
 sub found { $_[0]->{found} }
-sub pmc    { $_[0]->{pmc} }
+sub pmc   { $_[0]->{pmc} }
 sub open  { $_[0]->{open} }
 
 {
   package Module::Reader::File;
-  use constant _OPEN_STRING => "$]" >= 5.008 || (require IO::String, 0);
+  use constant _OPEN_STRING => "$]" >= 5.008 || !require IO::String;
   use Carp 'croak';
 
   sub new {
@@ -264,6 +294,7 @@ sub open  { $_[0]->{open} }
     $_[0]->{raw_filehandle} ||= !$_[0]->{disk_file} ? undef : do {
       open my $fh, '<'.Module::Reader::_OPEN_LAYERS, $_[0]->{disk_file}
         or croak "Can't locate $_[0]->{disk_file}";
+      $fh;
     };
   }
 
@@ -273,20 +304,22 @@ sub open  { $_[0]->{open} }
       if exists $self->{content};
     my $fh = $self->raw_filehandle;
     my $cb = $self->read_callback;
+    my $content = defined $self->{prefix} ? ${$self->{prefix}} : '';
     if ($fh && !$cb) {
       local $/;
-      return scalar <$fh>;
+      $content .= <$fh>;
     }
-    my @params = @{$self->read_callback_options||[]};
-    my $content = '';
-    while (1) {
-      local $_ = $fh ? <$fh> : '';
-      $_ = ''
-        if !defined;
-      # perlfunc/require says that the first parameter will be a reference the
-      # sub itself.  this is wrong.  0 will be passed.
-      last if !$cb->(0, @params);
-      $content .= $_;
+    if ($cb) {
+      my @params = @{$self->read_callback_options||[]};
+      while (1) {
+        local $_ = $fh ? <$fh> : '';
+        $_ = ''
+          if !defined;
+        # perlfunc/require says that the first parameter will be a reference the
+        # sub itself.  this is wrong.  0 will be passed.
+        last if !$cb->(0, @params);
+        $content .= $_;
+      }
     }
     return $self->{content} = $content;
   }
@@ -356,7 +389,7 @@ Returns an IO handle for the given module.
 
 Returns the content of a given module.
 
-=head1 CLASS ATTRIBUTES
+=head1 ATTRIBUTES
 
 =over 4
 
@@ -384,6 +417,19 @@ files.  If not specified, the same behavior perl was compiled with will be used.
 A boolean controlling if the files found will be opened immediately when found.
 Defaults to true.
 
+=item abort_on_eacces
+
+A boolean controlling if an error should be thrown or if the path should be
+skipped when encountering C<EACCES> (access denied) errors.  Defaults to true
+on perl 5.18 and above, matching the behavior of L<require|perlfunc/require>.
+
+=item check_hooks_for_nonsearchable
+
+For non-searchable paths (absolute paths and those starting with C<./> or
+C<../>) attempt to check the hook items (and not the directories) in C<@INC> if
+the file cannot be found directly.  This matches the behavior of perl.  Defaults
+to true.
+
 =back
 
 =head1 METHODS
@@ -421,7 +467,7 @@ memory or have arbitrary filters applied.
 
 =head3 filename
 
-The filename that was seached for.
+The filename that was searched for.
 
 =head3 module
 
@@ -532,7 +578,7 @@ files that it finds.
 
 =item L<Module::Locate>
 
-Innacurately searches C<@INC> for matching files.  Attempts to handle hooks, but
+Inaccurately searches C<@INC> for matching files.  Attempts to handle hooks, but
 handles most cases wrong.
 
 =item L<Module::Mapper>
